from .BaseTranslator import BaseTranslator
from ..QueryObject import QueryObject
from ..Builder import SelectBuilder, InsertBuilder, UpdateBuilder, DeleteBuilder

class GenericTranslator(BaseTranslator) :

    def __init__(self, query: QueryObject) :
        query.setMarkQuote("?", ":", "'")
        self.quote_struct = ""
        self.quote_string = "'"
        self.equal = "="
        self.open_bracket = "("
        self.close_bracket = ")"
        self.dot = "."
        self.comma = ", "
        self.end_query = ""

    def translateSelect(self, query: QueryObject, builder: SelectBuilder) :
        self.firstKeyword(query, builder.builderType())
        self.columnList(query, builder.getColumns(), builder.countColumns())
        self.fromTable(query, builder.getTable())
        self.where(query, builder.getWhere(), builder.countWhere())
        self.groupBy(query, builder.getGroup(), builder.countGroup())
        self.having(query, builder.getHaving(), builder.countHaving())
        self.orderBy(query, builder.getOrder(), builder.countOrder())
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())

    def translateInsert(self, query: QueryObject, builder: InsertBuilder) :
        self.firstKeyword(query, builder.builderType())
        self.intoTable(query, builder.getTable())
        self.columnListInsert(query, builder.getValues(), builder.countValues())
        self.valuesInsert(query, builder.getValues(), builder.countValues())
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())

    def translateUpdate(self, query: QueryObject, builder: UpdateBuilder) :
        self.firstKeyword(query, builder.builderType())
        self.tableSet(query, builder.getTable())
        self.valuesUpdate(query, builder.getValues(), builder.countValues())
        self.where(query, builder.getWhere(), builder.countWhere())
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())

    def translateDelete(self, query: QueryObject, builder: DeleteBuilder) :
        self.firstKeyword(query, builder.builderType())
        self.fromTable(query, builder.getTable())
        self.where(query, builder.getWhere(), builder.countWhere())
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())



from .QueryObject import QueryObject
from .Query import BaseQuery
from .Builder import SelectBuilder, InsertBuilder, UpdateBuilder, DeleteBuilder
from .Translator import BaseTranslator, GenericTranslator, MySQLTranslator

class QueryTranslator :

    TRANSLATOR_GENERIC = 1
    TRANSLATOR_BEAUTIFY = 2
    TRANSLATOR_MYSQL = 3
    TRANSLATOR_SQLITE = 4

    NO_PARAM = 1
    PARAM_NUM = 2
    PARAM_ASSOC = 3

    def __init__(self, translator: int = TRANSLATOR_GENERIC, bindingOption: int = PARAM_NUM) :
        self.queryObject = QueryObject()
        self.translator = translator
        self.bindingOption = bindingOption

    def getQueryObject(self) :
        return self.queryObject

    def translate(self, queryBuilder: BaseQuery, translator: int = 0, bindingOption: int = 0) -> str :
        if translator == 0: translator = self.translator
        if bindingOption == 0: bindingOption = self.bindingOption
        self.translateBuilder(self.queryObject, queryBuilder.getBuilder(), translator)
        return self.query(bindingOption)

    @staticmethod
    def translateBuilder(query: QueryObject, builder, translator: int) :
        translatorClass = QueryTranslator.getTranslator(query, translator)
        if isinstance(builder, SelectBuilder) :
            translatorClass.translateSelect(query, builder)
        elif isinstance(builder, InsertBuilder) :
            translatorClass.translateInsert(query, builder)
        elif isinstance(builder, UpdateBuilder) :
            translatorClass.translateUpdate(query, builder)
        elif isinstance(builder, DeleteBuilder) :
            translatorClass.translateDelete(query, builder)
        else :
            raise Exception('Tried to translate unregistered builder object')

    @staticmethod
    def getTranslator(query: QueryObject, translator: int) :
        if translator == QueryTranslator.TRANSLATOR_GENERIC :
            return GenericTranslator(query)
        elif translator == QueryTranslator.TRANSLATOR_MYSQL :
            return MySQLTranslator(query)
        else :
            raise Exception('Translator selected is not registered')

    @staticmethod
    def getBindingOption(bindingOption: int) :
        bindingFlag = False
        bindingMode = False
        if bindingOption == QueryTranslator.PARAM_ASSOC :
            bindingFlag = True
            bindingMode = True
        elif bindingOption == QueryTranslator.PARAM_NUM :
            bindingFlag = True
            bindingMode = False
        return (bindingFlag, bindingMode)

    def query(self, bindingOption: int = 0) -> str :
        if bindingOption == 0: bindingOption = self.bindingOption
        return self.getQuery(self.queryObject, bindingOption)

    @staticmethod
    def getQuery(query: QueryObject, bindingOption: int) -> str :
        (bindingFlag, bindingMode) = QueryTranslator.getBindingOption(bindingOption)
        queryString = ''
        parts = query.parts()
        params = query.params()
        if bindingMode : mark = query.bindMarkAssoc()
        else : mark = query.bindMarkNum()
        quote = query.stringQuote()
        for i in range(len(parts)) :
            queryString += parts[i]
            if i < len(params) :
                if bindingFlag :
                    if bindingMode : queryString += (mark + 'v' + str(i))
                    else : queryString += mark
                else :
                    if isinstance(params[i], str) : queryString += (quote + params[i] + quote)
                    else : queryString += str(params[i])
        return queryString

    def params(self, bindingOption: int = 0) :
        if bindingOption == 0: bindingOption = self.bindingOption
        return self.getParams(self.queryObject, bindingOption)

    @staticmethod
    def getParams(query: QueryObject, bindingOption: int) :
        (bindingFlag, bindingMode) = QueryTranslator.getBindingOption(bindingOption)
        array = None
        if bindingFlag :
            if bindingMode :
                array = {}
                for i, param in enumerate(query.params()) :
                    array['v' + str(i)] = param
            else :
                array = query.params()
        return array



from .BaseQuery import BaseQuery
from .Manipulation import Manipulation
from ..Builder import BaseBuilder, SelectBuilder

class Select(BaseQuery) :

    man = Manipulation()

    def __init__(self, options: tuple = (), statement = None) :
        self.builder = SelectBuilder()
        self.builder.builderType(BaseBuilder.SELECT)
        self.options = options
        self.statement = statement

    def select(self, table) :
        if table :
            tableObject = self.man.createTable(table)
            self.builder.setTable(tableObject)
        else :
            raise Exception("Table name is not defined")
        return self

    def column(self, column) :
        columnObject = self.man.createColumn(column)
        self.builder.addColumn(columnObject)
        return self

    def columns(self, columns) :
        columnObjects = self.man.createColumns(columns)
        for col in columnObjects :
            self.builder.addColumn(col)
        return self

    def beginWhere(self) :
        self.man.beginClause()
        return self

    def beginAndWhere(self) :
        self.man.beginAndClause()
        return self

    def beginOrWhere(self) :
        self.man.beginOrClause()
        return self

    def beginNotAndWhere(self) :
        self.man.beginNotAndClause()
        return self

    def beginNotOrWhere(self) :
        self.man.beginNotOrClause()
        return self

    def endWhere(self) :
        self.man.endClause(self.man.CLAUSE_WHERE, self.builder)
        return self

    def where(self, column, operator: str, value = None) :
        clauseObject = self.man.andClause(self.man.CLAUSE_WHERE, column, operator, value)
        self.builder.addWhere(clauseObject)
        return self

    def andWhere(self, column, operator: str, value = None) :
        clauseObject = self.man.andClause(self.man.CLAUSE_WHERE, column, operator, value)
        self.builder.addWhere(clauseObject)
        return self

    def orWhere(self, column, operator: str, value = None) :
        clauseObject = self.man.orClause(self.man.CLAUSE_WHERE, column, operator, value)
        self.builder.addWhere(clauseObject)
        return self

    def notAndWhere(self, column, operator: str, value = None) :
        clauseObject = self.man.notAndClause(self.man.CLAUSE_WHERE, column, operator, value)
        self.builder.addWhere(clauseObject)
        return self

    def notOrWhere(self, column, operator: str, value = None) :
        clauseObject = self.man.notOrClause(self.man.CLAUSE_WHERE, column, operator, value)
        self.builder.addWhere(clauseObject)
        return self

    def beginHaving(self) :
        self.man.beginClause()
        return self

    def beginAndHaving(self) :
        self.man.beginAndClause()
        return self

    def beginOrHaving(self) :
        self.man.beginOrClause()
        return self

    def beginNotAndHaving(self) :
        self.man.beginNotAndClause()
        return self

    def beginNotOrHaving(self) :
        self.man.beginNotOrClause()
        return self

    def endHaving(self) :
        self.man.endClause(self.man.CLAUSE_HAVING, self.builder)
        return self

    def having(self, column, operator: str, value = None) :
        clauseObject = self.man.andClause(self.man.CLAUSE_HAVING, column, operator, value)
        self.builder.addHaving(clauseObject)
        return self

    def andHaving(self, column, operator: str, value = None) :
        clauseObject = self.man.andClause(self.man.CLAUSE_HAVING, column, operator, value)
        self.builder.addHaving(clauseObject)
        return self

    def orHaving(self, column, operator: str, value = None) :
        clauseObject = self.man.orClause(self.man.CLAUSE_HAVING, column, operator, value)
        self.builder.addHaving(clauseObject)
        return self

    def notAndHaving(self, column, operator: str, value = None) :
        clauseObject = self.man.notAndClause(self.man.CLAUSE_HAVING, column, operator, value)
        self.builder.addHaving(clauseObject)
        return self

    def notOrHaving(self, column, operator: str, value = None) :
        clauseObject = self.man.notOrClause(self.man.CLAUSE_HAVING, column, operator, value)
        self.builder.addHaving(clauseObject)
        return self

    def groupBy(self, columns) :
        columnObjects = self.man.createGroups(columns)
        for col in columnObjects :
            self.builder.addGroup(col)
        return self

    def orderBy(self, columns, orderType) :
        orderObjects = self.man.createOrders(columns, orderType)
        for order in orderObjects :
            self.builder.addOrder(order)
        return self

    def orderByAsc(self, column) :
        orderObjects = self.man.orderAsc(column)
        for order in orderObjects :
            self.builder.addOrder(order)
        return self

    def orderByDesc(self, column) :
        orderObjects = self.man.orderDesc(column)
        for order in orderObjects :
            self.builder.addOrder(order)
        return self

    def limit(self, limit, offset = None) :
        limitObject = self.man.createLimit(limit, offset)
        self.builder.setLimit(limitObject)
        return self

    def offset(self, offset) :
        limitObject = self.man.offset(offset)
        self.builder.setLimit(limitObject)
        return self



from ..QueryObject import QueryObject
from ..QueryTranslator import QueryTranslator
from ..Builder import BaseBuilder

class BaseQuery :

    builder = BaseBuilder()
    queryObject = QueryObject()

    options = (QueryTranslator.TRANSLATOR_GENERIC, QueryTranslator.PARAM_NUM)
    statement = None
    translateFlag = True

    def getQueryObject(self) :
        return self.queryObject

    def getBuilder(self) :
        return self.builder

    def translate(self, translator: int = 0) :
        if translator == 0 : translator = self.options[0]
        self.queryObject = QueryObject()
        QueryTranslator.translateBuilder(self.queryObject, self.builder, translator)

    def query(self, translator: int = 0, bindingOption: int = 0) :
        if self.translateFlag :
            self.translate(translator)
            self.translateFlag = False
        if bindingOption == 0 : bindingOption = self.options[1]
        return QueryTranslator.getQuery(self.queryObject, bindingOption)

    def params(self, translator: int = 0, bindingOption: int = 0) :
        if self.translateFlag :
            self.translate(translator)
            self.translateFlag = False
        if bindingOption == 0 : bindingOption = self.options[1]
        return QueryTranslator.getParams(self.queryObject, bindingOption)



from .BaseQuery import BaseQuery
from .Manipulation import Manipulation
from ..Builder import BaseBuilder, InsertBuilder

class Insert(BaseQuery) :

    man = Manipulation()

    def __init__(self, options: tuple = (), statement = None) :
        self.builder = InsertBuilder()
        self.builder.builderType(BaseBuilder.INSERT)
        self.options = options
        self.statement = statement

    def insert(self, table) :
        if table :
            tableObject = self.man.createTable(table)
            self.builder.setTable(tableObject)
        else :
            raise Exception("Table name is not defined")
        return self

    def values(self, values) :
        valueObject = self.man.createValue(values)
        self.builder.addValue(valueObject)
        return self

    def multiValues(self, multiValues) :
        valuesObjects = self.man.createMultiValue(multiValues)
        for val in valuesObjects :
            self.builder.addValue(val)
        return self

    def limit(self, limit, offset = None) :
        limitObject = self.man.createLimit(limit, offset)
        self.builder.setLimit(limitObject)
        return self

    def offset(self, offset) :
        limitObject = self.man.offset(offset)
        self.builder.setLimit(limitObject)
        return self



from ..Builder import Table, Column, Value, Clause, Order, Limit
from typing import Iterable, Mapping

class Manipulation :

### TABLE, COLUMN, VALUES QUERY ###

    table = ''

    def createTable(self, table) -> Table :
        name = ''
        alias = ''
        if isinstance(table, str) :
            self.table = table
            name = table
            return Table(table)
        elif isinstance(table, Mapping) :
            if len(table) == 1 :
                alias = tuple(table.keys())[0]
                name = table[alias]
                self.table = str(alias)
        return Table(str(name), str(alias))

    def createColumn(self, column) -> Column :
        table = ''
        name = ''
        function = ''
        alias = ''
        if isinstance(column, str) :
            (table, name, function) = self.parseColumnStr(column)
        elif isinstance(column, Mapping) :
            (table, name, function, alias) = self.parseColumnMap(column)
        return Column(table, name, function, alias)

    def createColumns(self, columns) -> tuple :
        columnObjects = ()
        if isinstance(columns, Mapping) :
            keys = tuple(columns.keys())
            for i in range(len(keys)) :
                columnObjects += (self.createColumn({keys[i]: columns[keys[i]]}),)
        elif isinstance(columns, Iterable) :
            for col in columns :
                columnObjects += (self.createColumn(col),)
        return columnObjects

    def parseColumnStr(self, column: str) -> tuple :
        function = ''
        pos1 = column.find('(')
        pos2 = column.rfind(')')
        if pos1 > 0 and pos2 > 1 :
            function = column[0:pos1]
            column = column[pos1+1:pos2]
        table = self.table
        name = column
        split = column.split('.')
        if len(split) == 2 :
            table = split[0]
            name = split[1]
        return (table, name, function)

    def parseColumnMap(self, column: Mapping) -> tuple :
        alias = ''
        keys = tuple(column.keys())
        if len(column) == 1 :
            alias = str(keys[0])
            column = column[keys[0]]
        table = self.table
        name = ''
        function = ''
        if isinstance(column, str) :
            (table, name, function) = self.parseColumnStr(column)
        return (table, name, function, alias)

    def createValue(self, inputValue) -> Value :
        columns = ()
        values = ()
        if isinstance(inputValue, Mapping) :
            columns = tuple(inputValue.keys())
            values = tuple(inputValue.values())
        elif isinstance(inputValue, Iterable) :
            (columns, values) = self.parseValuePair(inputValue)
        return Value(self.table, columns, values)

    def createMultiValue(self, multiValues) -> tuple :
        valuesObjects = ()
        if isinstance(multiValues, Iterable) :
            for val in multiValues :
                valuesObjects += (self.createValue(val),)
        return valuesObjects

    def parseValuePair(self, pairs: Iterable) -> tuple :
        if isinstance(pairs[0], str) and len(pairs) == 2 :
            return ((pairs[0],), (pairs[1],))
        columns = ()
        values = ()
        for pair in pairs :
            if isinstance(pair, Mapping) :
                if len(pair) :
                    key = tuple(pair.keys())[0]
                    columns += (key,)
                    values += (pair[key],)
            elif isinstance(pair, Iterable) :
                if len(pair) == 2 :
                    columns += (pair[0],)
                    values += (pair[1],)
        return (columns, values)

### WHERE AND HAVING CLAUSE QUERY ###

    CLAUSE_DEFAULT = 0
    CLAUSE_WHERE = 1
    CLAUSE_HAVING = 2

    clauseType = CLAUSE_DEFAULT
    nestedConjunctive = Clause.CONJUNCTIVE_NONE
    nestedLevel = 0

    def createClause(self, clauseType: int, column, operator, values, conjunctive: int) -> Clause :
        columnObject = self.createColumn(column)
        validOperator = self.getOperator(operator)
        validValues = self.getValues(values, operator)
        conjunctive = self.getConjunctive(clauseType, conjunctive)
        nestedLevel = self.nestedLevel
        self.clauseType = clauseType
        self.nestedLevel = 0
        return Clause(columnObject, validOperator, validValues, conjunctive, nestedLevel)

    def getOperator(self, operator) -> int :
        if isinstance(operator, int) :
            validOperator = operator
        else :
            if operator == '=' or operator == '==' :
                validOperator = Clause.OPERATOR_EQUAL
            elif operator == '!=' or operator == '<>' :
                validOperator = Clause.OPERATOR_NOT_EQUAL
            elif operator == '>' :
                validOperator = Clause.OPERATOR_GREATER
            elif operator == '>=' :
                validOperator = Clause.OPERATOR_GREATER_EQUAL
            elif operator == '<' :
                validOperator = Clause.OPERATOR_LESS
            elif operator == '<=' :
                validOperator = Clause.OPERATOR_LESS_EQUAL
            elif operator == 'BETWEEN' :
                validOperator = Clause.OPERATOR_BETWEEN
            elif operator == 'NOT BETWEEN' :
                validOperator = Clause.OPERATOR_NOT_BETWEEN
            elif operator == 'LIKE' :
                validOperator = Clause.OPERATOR_LIKE
            elif operator == 'NOT LIKE' :
                validOperator = Clause.OPERATOR_NOT_LIKE
            elif operator == 'IN' :
                validOperator = Clause.OPERATOR_IN
            elif operator == 'NOT IN' :
                validOperator = Clause.OPERATOR_NOT_IN
            elif operator == 'NULL' or operator == 'IS NULL' :
                validOperator = Clause.OPERATOR_NULL
            elif operator == 'NOT NULL' or operator == 'IS NOT NULL' :
                validOperator = Clause.OPERATOR_NOT_NULL
            else :
                validOperator = Clause.OPERATOR_DEFAULT
        return validOperator
            
    def getValues(self, values, operator) :
        valid = True
        if operator == Clause.OPERATOR_BETWEEN or operator == Clause.OPERATOR_NOT_BETWEEN :
            if isinstance(values, Iterable) :
                valid = len(values) == 2
        if operator == Clause.OPERATOR_IN or operator == Clause.OPERATOR_NOT_IN :
            valid = isinstance(values, Iterable)
        if valid :
            return values
        else :
            raise Exception('Invalid input values for Where or Having clause')

    def getConjunctive(self, clauseType: int, conjunctive: int) -> int :
        if clauseType == self.clauseType :
            if conjunctive == Clause.CONJUNCTIVE_NONE :
                if self.nestedConjunctive == Clause.CONJUNCTIVE_NONE : return Clause.CONJUNCTIVE_AND
                else : return self.nestedConjunctive
            else :
                return conjunctive
        else :
            return Clause.CONJUNCTIVE_NONE

    def beginClause(self) :
        self.nestedConjunctive = Clause.CONJUNCTIVE_NONE
        self.nestedLevel -= 1

    def beginAndClause(self) :
        self.nestedConjunctive = Clause.CONJUNCTIVE_AND
        self.nestedLevel -= 1

    def beginOrClause(self) :
        self.nestedConjunctive = Clause.CONJUNCTIVE_OR
        self.nestedLevel -= 1

    def beginNotAndClause(self) :
        self.nestedConjunctive = Clause.CONJUNCTIVE_NOT_AND
        self.nestedLevel -= 1

    def beginNotOrClause(self) :
        self.nestedConjunctive = Clause.CONJUNCTIVE_NOT_OR
        self.nestedLevel -= 1

    def endClause(self, clauseType: int, builder) :
        if clauseType == self.CLAUSE_WHERE :
            lastNested = builder.lastWhere().nestedLevel()
            builder.editWhereNested(lastNested + 1)
        elif clauseType == self.CLAUSE_HAVING :
            lastNested = builder.lastHaving().nestedLevel()
            builder.editHavingNested(lastNested + 1)

    def andClause(self, clauseType: int, column, operator: str, value = None) -> Clause :
        return self.createClause(clauseType, column, operator, value, Clause.CONJUNCTIVE_AND)

    def orClause(self, clauseType: int, column, operator: str, value = None) -> Clause :
        return self.createClause(clauseType, column, operator, value, Clause.CONJUNCTIVE_OR)

    def notAndClause(self, clauseType: int, column, operator: str, value = None) -> Clause :
        return self.createClause(clauseType, column, operator, value, Clause.CONJUNCTIVE_NOT_AND)

    def notOrClause(self, clauseType: int, column, operator: str, value = None) -> Clause :
        return self.createClause(clauseType, column, operator, value, Clause.CONJUNCTIVE_NOT_OR)

### GROUP BY QUERY ###

    def createGroups(self, columns) -> tuple :
        if (isinstance(columns, Mapping) and len(columns) == 1) or isinstance(columns, str) :
            return (self.createColumn(columns),)
        else :
            return self.createColumns(columns)

### ORDER BY QUERY ###

    def createOrders(self, columns, orderType) -> tuple :
        if (isinstance(columns, Mapping) and len(columns) == 1) or isinstance(columns, str) :
            columnObjects = (self.createColumn(columns),)
        else :
            columnObjects = self.createColumns(columns)
        validType = self.getOrderType(orderType)
        orderObjects = ()
        for col in columnObjects :
            orderObjects += (Order(col, validType),)
        print(orderObjects)
        return orderObjects

    def getOrderType(self, orderType) -> int :
        if isinstance(orderType, int) :
            validType = orderType
        else :
            if orderType == 'ASCENDING' or orderType == 'ASC' or orderType == 'ascending' or orderType == 'asc' :
                validType = Order.ORDER_ASC
            elif orderType == 'DESCENDING' or orderType == 'DESC' or orderType == 'descending' or orderType == 'desc' :
                validType = Order.ORDER_DESC
            else :
                validType = Order.ORDER_NONE
        return validType

    def orderAsc(self, column) -> Order :
        return self.createOrders(column, Order.ORDER_ASC)

    def orderDesc(self, column) -> Order :
        return self.createOrders(column, Order.ORDER_DESC)

### LIMIT AND OFFSET QUERY ###

    def createLimit(self, limit, offset) -> Limit :
        validLimit = Limit.NOT_SET
        validOffset = Limit.NOT_SET
        if isinstance(limit, int) :
            if limit > 0: validLimit = limit
        if isinstance(offset, int) :
            if offset > 0: validOffset = offset
        return Limit(validLimit, validOffset)

    def offset(self, offset) -> Limit :
        return self.createLimit(Limit.NOT_SET, offset)



from .Query import BaseQuery, Select, Insert, Update, Delete
from .QueryTranslator import QueryTranslator

class QueryBuilder :

    table = ''

    options = ()
    statement = None

    def __init__(self, translator: int = QueryTranslator.TRANSLATOR_GENERIC, bindingOption: int = QueryTranslator.PARAM_NUM) :
        self.options = (translator, bindingOption)
        self.statement = None

    def setStatement(self, statement) :
        self.statement = statement

    def setTable(self, table: str) :
        self.table = table
        return self

    def select(self, table = None) -> Select :
        if table is None :
            table = self.table
        selectQuery = Select(self.options, self.statement)
        return selectQuery.select(table)

    def insert(self, table = None) -> Insert :
        if table is None :
            table = self.table
        insertQuery = Insert(self.options, self.statement)
        return insertQuery.insert(table)

    def update(self, table = None) -> Update :
        if table is None :
            table = self.table
        updateQuery = Update(self.options, self.statement)
        return updateQuery.update(table)

    def delete(self, table = None) -> Delete :
        if table is None :
            table = self.table
        deleteQuery = Delete(self.options, self.statement)
        return deleteQuery.delete(table)



class Table :

    def __init__(self, name: str, alias: str = '') :
        self.__name = name
        self.__alias = alias

    def name(self) -> str :
        return self.__name

    def alias(self) -> str :
        return self.__alias

class Column :

    def __init__(self, table: str, name: str, function: str = '', alias: str = '') :
        self.__table = table
        self.__name = name
        self.__function = function
        self.__alias = alias

    def table(self) -> str :
        return self.__table

    def name(self) -> str :
        return self.__name

    def function(self) -> str :
        return self.__function

    def alias(self) -> str :
        return self.__alias

class Value :

    def __init__(self, table: str, columns: tuple, values: tuple) :
        self.__table = table
        lenCol = len(columns)
        lenVal = len(values)
        lenMin = lenCol
        if lenCol > lenVal : lenMin = lenVal
        self.__columns = ()
        self.__values = values[:lenMin]
        for i in range(lenMin) :
            self.__columns += (str(columns[i]),)

    def table(self) -> str :
        return self.__table

    def columns(self) -> tuple :
        return self.__columns

    def values(self) -> tuple :
        return self.__values

class Clause :

    OPERATOR_DEFAULT = 0
    OPERATOR_EQUAL = 1
    OPERATOR_NOT_EQUAL = 2
    OPERATOR_GREATER = 3
    OPERATOR_GREATER_EQUAL = 4
    OPERATOR_LESS = 5
    OPERATOR_LESS_EQUAL = 6
    OPERATOR_LIKE = 7
    OPERATOR_NOT_LIKE = 8
    OPERATOR_BETWEEN = 9
    OPERATOR_NOT_BETWEEN = 10
    OPERATOR_IN = 11
    OPERATOR_NOT_IN = 12
    OPERATOR_NULL = 13
    OPERATOR_NOT_NULL = 14

    CONJUNCTIVE_NONE = 0
    CONJUNCTIVE_AND = 1
    CONJUNCTIVE_OR = 2
    CONJUNCTIVE_NOT_AND = 3
    CONJUNCTIVE_NOT_OR = 4

    def __init__(self, column, operator: int, value, conjunctive: int, nestedLevel: int) :
        self.__column = column
        if operator > 0 and operator <= 14 :
            self.__operator = operator
        else :
            self.__operator = self.OPERATOR_DEFAULT
        self.__value = value
        if conjunctive > 0 and conjunctive <= 4 :
            self.__conjunctive = conjunctive
        else :
            self.__conjunctive = self.CONJUNCTIVE_NONE
        self.__nestedLevel = nestedLevel

    def column(self) :
        return self.__column

    def operator(self) -> int :
        return self.__operator

    def value(self) :
        return self.__value

    def conjunctive(self) -> int :
        return self.__conjunctive

    def nestedLevel(self, input: int = -1) -> int :
        if input != -1 : self.__nestedLevel = input
        return self.__nestedLevel

class Order :

    ORDER_NONE = 0
    ORDER_ASC = 1
    ORDER_DESC = 2

    def __init__(self, column: Column, orderType: int) :
        self.__column = column
        self.__orderType = orderType

    def column(self) -> Column :
        return self.__column

    def orderType(self) -> int :
        return self.__orderType

class Limit :

    NOT_SET = -1

    def __init__(self, limit: int, offset: int) :
        self.__limit = limit
        self.__offset = offset

    def limit(self) -> int :
        return self.__limit

    def offset(self) -> int :
        return self.__offset



from .BaseBuilder import BaseBuilder
from .Structure import Limit

class InsertBuilder(BaseBuilder) :

    def __init__(self) :
        BaseBuilder.__init__(self)
        self.__limit = None

    def getLimit(self) -> tuple :
        return self.__limit

    def hasLimit(self) -> bool :
        if self.__limit is None : return False
        else : return True

    def setLimit(self, limit: Limit) :
        self.__limit = limit



from ..QueryObject import QueryObject
from ..Builder import BaseBuilder
from ..Builder import Table, Column, Value, Clause, Order, Limit

class BaseTranslator :

    def __init__(self) :
        self.quote_struct = "`"
        self.quote_string = "'"
        self.equal = "="
        self.open_bracket = "("
        self.close_bracket = ")"
        self.dot = "."
        self.comma = ","
        self.end_query = ""

    def firstKeyword(self, query: QueryObject, builderType: int) :
        if builderType == BaseBuilder.SELECT :
            query.add('SELECT ')
        elif builderType == BaseBuilder.INSERT :
            query.add('INSERT ')
        elif builderType == BaseBuilder.UPDATE :
            query.add('UPDATE ')
        elif builderType == BaseBuilder.DELETE :
            query.add('DELETE')

    def fromTable(self, query: QueryObject, table: Table) :
        name = table.name()
        alias = table.alias()
        query.add(' FROM ' + self.quote_struct)
        query.add(name)
        query.add(self.quote_struct)
        if alias :
            query.add(' AS ' + self.quote_string + alias + self.quote_string)

    def intoTable(self, query: QueryObject, table: Table) :
        name = table.name()
        query.add('INTO ' + self.quote_struct)
        query.add(name)
        query.add(self.quote_struct)

    def tableSet(self, query: QueryObject, table: Table) :
        name = table.name()
        query.add(self.quote_struct)
        query.add(name)
        query.add(self.quote_struct)

    def columnList(self, query: QueryObject, columns: tuple, count: int) :
        if count == 0 :
            query.add('*')
            return
        for column in columns :
            if isinstance(column, Column) :
                name = column.name()
                alias = column.alias()
                function = column.function()
                if function :
                    query.add(function + self.open_bracket)
                query.add(self.quote_struct)
                query.add(name)
                query.add(self.quote_struct)
                if function :
                    query.add(self.close_bracket)
                if alias :
                    query.add(' AS ' + self.quote_string + alias + self.quote_string)
                count -= 1
                if count > 0 : query.add(self.comma)

    def columnListInsert(self, query: QueryObject, values: tuple, count: int) :
        if count == 0 :
            query.add(' ' + self.open_bracket)
            query.add(self.close_bracket)
            return
        value = values[0]
        if isinstance(value, Value) :
            columns = value.columns()
            count = len(columns)
            query.add(' ' + self.open_bracket)
            for column in columns : 
                query.add(self.quote_struct)
                query.add(column)
                query.add(self.quote_struct)
                count -= 1
                if count > 0 : query.add(self.comma)
            query.add(self.close_bracket)

    def column(self, query: QueryObject, column: Column) :
        name = column.name()
        alias = column.alias()
        function = column.function()
        if function :
            query.add(function + self.open_bracket)
        if alias :
            query.add(self.quote_struct)
            query.add(column.alias())
            query.add(self.quote_struct)
        else :
            query.add(self.quote_struct)
            query.add(name)
            query.add(self.quote_struct)
        if function :
            query.add(self.close_bracket)

    def valuesInsert(self, query: QueryObject, values: tuple, count: int) :
        query.add(' VALUES ')
        if count == 0 :
            query.add(self.open_bracket)
            query.add(self.close_bracket)
            return
        for value in values :
            if isinstance(value, Value) :
                vals = value.values()
                countVals = len(vals)
                query.add(self.open_bracket)
                for val in vals :
                    query.add(val, True)
                    countVals -= 1
                    if countVals > 0 : query.add(self.comma)
                query.add(self.close_bracket)
            count -= 1
            if count > 0 : query.add(self.comma)

    def valuesUpdate(self, query: QueryObject, values: tuple, count: int) :
        query.add(' SET ')
        for value in values :
            if isinstance(value, Value) :
                columns = value.columns()
                vals = value.values()
                countVals = len(vals)
                for i, val in enumerate(vals) :
                    query.add(self.quote_struct)
                    query.add(columns[i])
                    query.add(self.quote_struct + self.equal)
                    query.add(val, True)
                    countVals -= 1
                    if countVals > 0 : query.add(self.comma)
            count -= 1
            if count > 1 : query.add(self.comma)

    def operator(self, operator: int) -> str :
        if operator == Clause.OPERATOR_EQUAL :
            return '='
        elif operator == Clause.OPERATOR_NOT_EQUAL :
            return '!='
        elif operator == Clause.OPERATOR_GREATER :
            return '>'
        elif operator == Clause.OPERATOR_GREATER_EQUAL :
            return '>='
        elif operator == Clause.OPERATOR_LESS :
            return '<'
        elif operator == Clause.OPERATOR_LESS_EQUAL :
            return '<='
        elif operator == Clause.OPERATOR_BETWEEN :
            return ' BETWEEN '
        elif operator == Clause.OPERATOR_NOT_BETWEEN :
            return ' NOT BETWEEN '
        elif operator == Clause.OPERATOR_LIKE :
            return ' LIKE '
        elif operator == Clause.OPERATOR_NOT_LIKE :
            return ' NOT LIKE '
        elif operator == Clause.OPERATOR_IN :
            return ' IN '
        elif operator == Clause.OPERATOR_NOT_IN :
            return ' NOT IN '
        elif operator == Clause.OPERATOR_NULL :
            return ' IS NULL '
        elif operator == Clause.OPERATOR_NOT_NULL :
            return ' IS NOT NULL '
        else :
            return ''

    def conjunctive(self, conjunctive: int) -> str :
        if conjunctive == Clause.CONJUNCTIVE_AND :
            return ' AND '
        elif conjunctive == Clause.CONJUNCTIVE_OR :
            return ' OR '
        elif conjunctive == Clause.CONJUNCTIVE_NOT_AND :
            return ' NOT AND '
        elif conjunctive == Clause.CONJUNCTIVE_NOT_OR :
            return ' NOT OR '
        else :
            return ''

    def brackets(self, level: int) -> str :
        string = ''
        if level < 0 :
            for i in range(level, 0) :
                string += self.open_bracket
        elif level > 0 :
            for i in range(0, level) :
                string += self.close_bracket
        return string

    def where(self, query: QueryObject, whereClauses: tuple, count: int) :
        if count :
            query.add(' WHERE ')
            for where in whereClauses :
                if isinstance(where, Clause) :
                    conjunctive = self.conjunctive(where.conjunctive())
                    nestedLevel = where.nestedLevel()
                    query.add(conjunctive)
                    if nestedLevel < 0 : query.add(self.brackets(nestedLevel))
                    self.clause(query, where)
                    if nestedLevel > 0 : query.add(self.brackets(nestedLevel))

    def having(self, query: QueryObject, havingClauses: tuple, count: int) :
        if count :
            query.add(' HAVING ')
            for having in havingClauses :
                if isinstance(having, Clause) :
                    conjunctive = self.conjunctive(having.conjunctive())
                    nestedLevel = having.nestedLevel()
                    query.add(conjunctive)
                    if nestedLevel < 0 : query.add(self.brackets(nestedLevel))
                    self.clause(query, having)
                    if nestedLevel > 0 : query.add(self.brackets(nestedLevel))

    def clause(self, query: QueryObject, clause: Clause) :
        column = clause.column()
        operator = clause.operator()
        value = clause.value()
        if isinstance(column, Column) :
            self.column(query, column)
        if operator == Clause.OPERATOR_BETWEEN :
            self.clauseBetween(query, value)
        elif operator == Clause.OPERATOR_IN :
            self.clauseIn(query, value)
        else :
            self.clauseComparison(query, value, operator)

    def clauseComparison(self, query: QueryObject, value, operator: int) :
        query.add(self.operator(operator))
        query.add(value, True)

    def clauseBetween(self, query: QueryObject, value) :
        query.add(self.operator(Clause.OPERATOR_BETWEEN))
        query.add(value[0], True)
        query.add(' AND ')
        query.add(value[1], True)

    def clauseIn(self, query: QueryObject, value) :
        query.add(self.operator(Clause.OPERATOR_IN))
        query.add(self.open_bracket)
        count = len(value)
        for i in range(count) :
            query.add(value[i], True)
            if i < count - 1 : query.add(self.comma)
        query.add(self.close_bracket)

    def groupBy(self, query: QueryObject, groups: tuple, count: int) :
        if count :
            query.add(' GROUP BY ')
            for group in groups :
                if isinstance(group, Column) :
                    self.column(query, group)
                    count -= 1
                    if count > 0 : query.add(self.comma)

    def orderBy(self, query: QueryObject, orders: tuple, count: int) :
        if count :
            query.add(' ORDER BY ')
            for order in orders :
                if isinstance(order, Order) :
                    self.column(query, order.column())
                    if (order.orderType() == Order.ORDER_ASC) :
                        query.add(' ASC')
                    elif (order.orderType() == Order.ORDER_DESC) :
                        query.add(' DESC')
                    count -= 1
                    if count > 0 : query.add(self.comma)

    def limitOffset(self, query: QueryObject, limitOffset: Limit, flag: bool) :
        if flag :
            limit = limitOffset.limit()
            offset = limitOffset.offset()
            if limit == Limit.NOT_SET :
                query.add(' OFFSET ')
                query.add(offset, True)
            else :
                query.add(' LIMIT ')
                query.add(limit, True)
                if offset != Limit.NOT_SET :
                    query.add(self.comma)
                    query.add(offset, True)



from .Structure import Table, Column, Value, Clause, Order, Limit
from .BaseBuilder import BaseBuilder
from .SelectBuilder import SelectBuilder
from .InsertBuilder import InsertBuilder
from .UpdateBuilder import UpdateBuilder
from .DeleteBuilder import DeleteBuilder


from .BaseBuilder import BaseBuilder
from .Structure import Clause, Limit

class UpdateBuilder(BaseBuilder) :

    def __init__(self) :
        BaseBuilder.__init__(self)
        self.__where = ()
        self.__limit = None

    def getWhere(self) -> tuple:
        return self.__where

    def lastWhere(self) -> Clause :
        count = len(self.__where)
        if count > 0 :
            return self.__where[count-1]
        return None

    def countWhere(self) -> int :
        return len(self.__where)

    def addWhere(self, where: Clause) :
        self.__where += (where,)

    def editWhereNested(self, nestedLevel: int) :
        count = len(self.__where)
        if count > 0 :
            self.__where[count-1].nestedLevel(nestedLevel)

    def getLimit(self) -> tuple :
        return self.__limit

    def hasLimit(self) -> bool :
        if self.__limit is None : return False
        else : return True

    def setLimit(self, limit: Limit) :
        self.__limit = limit
